//********************************************************************
UNIT Terrain;
// Note: these routines need a valid rendering context
//********************************************************************
INTERFACE

USES Windows,Classes,SysUtils,dglOpenGl;
//********************************************************************

Procedure InitOpenGlParams;
Function LoadTGA_8Bit_256X256_HeightMap(AFileName: String): Boolean;
//Function LoadTGA_24Bit_256X256_Texture(AFileName: String): Boolean;
Procedure FreeTexture;
Procedure RenderByOpenGl(ADeviceContext: Cardinal);

//********************************************************************
IMPLEMENTATION
//********************************************************************
CONST
   SIDE = 256;
   HSCALE = 0.1;

TYPE
   T8bPixel = Byte;
   T8bPixelMap = Array [0..SIDE*SIDE-1] Of T8bPixel;

   TVertex = Packed Record TexX,TexY,X,Y,Z: Single; End;
   TBiNormal = Array [1..2] Of TVertex;
   THeightMap = Array [0..SIDE*SIDE-1] Of TVertex;
   TNormalMap = Array [0..SIDE*SIDE-1] Of TBiNormal;
//********************************************************************
{All Targa formats are identified by a Data Type field, which is a one
byte binary integer located in byte three of the file.  The various
file types specified by this field are as follows:

       0  -  No image data included.
       1  -  Uncompressed, color-mapped (GreyScale?) images.
       2  -  Uncompressed, RGB images.
       3  -  Uncompressed, black and white images.
       9  -  Runlength encoded color-mapped images.
      10  -  Runlength encoded RGB images.
      11  -  Compressed, black and white images.
      32  -  Compressed color-mapped data, using Huffman, Delta, and
                    runlength encoding.
      33  -  Compressed color-mapped data, using Huffman, Delta, and
                    runlength encoding.  4-pass quadtree-type process.
}
   TTargaHeader = Packed Record  // For Unmapped RGB images (ImageType=2)
      ImageIdentLen: Byte;  // Length of the Image Identification Field (see below)
      ColorMapType: Byte;   // Color Map Type (0=NoColorMap,1=Colormap exists)
      ImageDataType: Byte;  // See description above
      ColMapSpec: Array[0..4] of Byte; // Ignored: Color Map Specification

      OrigX: Word;   // X coordinate of the lower left corner
      OrigY: Word;   // Y coordinate of the lower left corner
      tgWidth: Word;   // Width of the image in pixels
      tgHeight: Word;  // Height of the image in pixels
      tgBitsPerPix: Byte;  // Number of bits in a pixel: 16/24/32
      ImageDescr: Byte;  // Image Descriptor Byte
      // ImageIdent  // Image Identification Field, usually omitted
   End;
//********************************************************************
VAR
   XRange,YRange,RotAngle: Integer;
   HeightMap: THeightMap;
   NormalMap: TNormalMap;
   Texture: Cardinal;
//********************************************************************
// Creates the direction vector FromPoint->ToPoint
Function Vector3D(FromPoint,ToPoint: TVertex): TVertex;
Begin
   Result.X:= ToPoint.X - FromPoint.X;
   Result.Y:= ToPoint.Y - FromPoint.Y;
   Result.Z:= ToPoint.Z - FromPoint.Z;
End;
//********************************************************************
Function VectorNormalize(AVector: TVertex): TVertex;
Const Tolerance = 0.01;
Var LVectorValue: Single;
Begin
   With AVector Do LVectorValue:= Sqrt(X*X+Y*Y+Z*Z);
   If LVectorValue > Tolerance
      Then With Result Do Begin
         X:= AVector.X / LVectorValue;
         Y:= AVector.Y / LVectorValue;
         Z:= AVector.Z / LVectorValue;
      End
      Else With Result Do Begin
         X:=0; Y:=0; Z:=0;
      End;
End;
//********************************************************************
// Calculates the cross product vector (= a vector, that is orthogonal
// to both of them)
Function VectorCrossProduct(Vector1,Vector2: TVertex): TVertex;
Begin
   Result.X:= Vector1.Y*Vector2.Z - Vector1.Z*Vector2.Y;
   Result.Y:= Vector1.Z*Vector2.X - Vector1.X*Vector2.Z;
   Result.Z:= Vector1.X*Vector2.Y - Vector1.Y*Vector2.X;

   Result:= VectorNormalize(Result);
End;
//********************************************************************
// Loads 8bit uncompressed TGA (GreyScale image), where Width=Height=256
Function LoadTGA_8Bit_256X256_HeightMap(AFileName: String): Boolean;
Var
   FileHeader: TTargaHeader;
   LocalStream: TFileStream;
   PixelMap: T8bPixelMap;
   CountBytes,CountPixel,MapSize,Col,Row: Integer;
Begin
   Result:= False;

   If FileExists(AFileName) Then Begin
      LocalStream:= TFileStream.Create(AFileName,
                                fmOpenRead,fmShareDenyWrite);
      FillChar(FileHeader, SizeOf(FileHeader), #0);
      LocalStream.Read(FileHeader, SizeOf(FileHeader));
      With FileHeader Do
         If (tgBitsPerPix = 8) And (ImageDataType = 3) And
            (tgWidth = SIDE) And (tgHeight = SIDE) Then Begin

            MapSize:= SizeOf(PixelMap);
            CountBytes:= LocalStream.Read(PixelMap,MapSize);

            // Generate vertices
            If CountBytes = MapSize Then Begin
               For Row:= 0 To SIDE-1 Do
                  For Col:= 0 To SIDE-1 Do
                     With HeightMap[SIDE*Row+Col] Do Begin
                        // HeightValue
                        If (Col > 10) And (Col < SIDE-10) And
                           (Row > 10) And (Row < SIDE-10)
                           Then Y:= Trunc(PixelMap[SIDE*Row+Col]*HSCALE)
                           Else Y:= 0;
                        // Col/Row-values + Tex coords
                        With HeightMap[SIDE*Row+Col] Do Begin
                           TexX:= Col/SIDE;
                           TexY:= Row/SIDE;
                           X:= Col - SIDE Div 2;
                           Z:= Row - SIDE Div 2;
                        End;
                     End;

               // Calc normals
               For Row:= 0 To SIDE-2 Do
                  For Col:= 0 To SIDE-2 Do Begin
                     NormalMap[SIDE*Row+Col][1]:= VectorCrossProduct(
                        Vector3D(HeightMap[SIDE*(Row+1)+Col],HeightMap[SIDE*(Row+1)+Col+1]),
                        Vector3D(HeightMap[SIDE*(Row+1)+Col],HeightMap[SIDE*Row+Col+1]));
                     NormalMap[SIDE*Row+Col][2]:= VectorCrossProduct(
                        Vector3D(HeightMap[SIDE*(Row+1)+Col],HeightMap[SIDE*Row+Col+1]),
                        Vector3D(HeightMap[SIDE*(Row+1)+Col],HeightMap[SIDE*Row+Col]));
                  End;

               // Generate texture
               glGenTextures(1, @Texture);
               glBindTexture(GL_TEXTURE_2D, Texture);
               glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
               glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
               glTexImage2D(GL_TEXTURE_2D,
                     0,
                     GL_RGB,
                     SIDE,SIDE,
                     0,
                     GL_Luminance,
                     GL_UNSIGNED_BYTE,
                     @PixelMap);
               glBindTexture(GL_TEXTURE_2D, 0);

               Result:= True;
            End;

         End;

      LocalStream.Free;
   End;
End;
//********************************************************************
Procedure FreeTexture;
Begin
   glDeleteTextures(1,@Texture);
End;
//********************************************************************
Procedure InitOpenGlParams;
Const
   Light_Position: Array[0..3] Of GlFloat = (-10,+10,+10,0);
   Light_Ambient : Array[0..3] Of GlFloat = (0.4, 0.4, 0.4, 1.0);
   Light_Diffuse : Array[0..3] Of GlFloat = (0.8, 0.8, 0.8, 1.0);
   Mat_Ambient   : Array[0..3] Of GlFloat = (0.4, 0.4, 0.4, 1.0);
   Mat_Diffuse   : Array[0..3] Of GlFloat = (0.8, 0.8, 0.8, 1.0);

Var Viewport: Array [0..3] Of Integer;
Begin
   glEnable(GL_DEPTH_TEST);
   glDepthFunc(GL_LEQUAL);

   glEnable(GL_CULL_FACE);
   glCullFace(GL_BACK);

   glEnable(GL_Texture_2D);

   glDisable(GL_BLEND);

   glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);

   glShadeModel(GL_SMOOTH);

   glLightfv(GL_LIGHT0,GL_AMBIENT,@Light_Ambient);
   glLightfv(GL_LIGHT0,GL_DIFFUSE,@Light_Diffuse);
   glLightfv(GL_LIGHT0,GL_POSITION,@Light_Position);
   glEnable(GL_LIGHTING);
   glEnable(GL_LIGHT0);

   glMaterialfv(GL_FRONT, GL_AMBIENT,@Mat_Ambient);
   glMaterialfv(GL_FRONT, GL_DIFFUSE,@Mat_Diffuse);
   glEnable(GL_COLOR_MATERIAL);

   glClearColor(0,0,0,1); // BLACK

   glGetIntegerv(GL_VIEWPORT, @Viewport);
   XRange:= ViewPort[2];
   YRange:= ViewPort[3];
   Rotangle:= 0;
End;
//********************************************************************
Procedure RenderTriangle(ANormal,A,B,C: TVertex);
Begin
   With ANormal Do glNormal3f(X,Y,Z);
   With A Do Begin
      glTexCoord2f(TexX,TexY);
      glVertex3f(X,Y,Z);
   End;
   With B Do Begin
      glTexCoord2f(TexX,TexY);
      glVertex3f(X,Y,Z);
   End;
   With C Do Begin
      glTexCoord2f(TexX,TexY);
      glVertex3f(X,Y,Z);
   End;
End;
//********************************************************************
Procedure RenderHeightMap;
Var
   Col,Row: Integer;
Begin
   glBegin(GL_TRIANGLES);
   For Row:= 0 To SIDE-2 Do
      For Col:= 0 To SIDE-2 Do Begin
         RenderTriangle(NormalMap[SIDE*Row+Col][1],
                        HeightMap[SIDE*(Row+1)+Col],   // Lower left
                        HeightMap[SIDE*(Row+1)+Col+1], // Lower right
                        HeightMap[SIDE*Row+Col+1]);    // Upper right
         RenderTriangle(NormalMap[SIDE*Row+Col][2],
                        HeightMap[SIDE*(Row+1)+Col],   // Lower left
                        HeightMap[SIDE*Row+Col+1],     // Upper right
                        HeightMap[SIDE*Row+Col]);      // Upper left
      End;
   glEnd;
End;
//********************************************************************
Procedure RenderByOpenGl(ADeviceContext: Cardinal);
Var
   XMax,YMax,ZNear,ZFar: Single;
   ErrorString: String;
Begin
   glClear(GL_COLOR_BUFFER_BIT Or GL_DEPTH_BUFFER_BIT);

   glMatrixMode(GL_PROJECTION);
   glLoadIdentity;
   ZNear:= +1;
   ZFar := XRange*1.1;
   gluPerspective(45.0,XRange/YRange,ZNear,ZFar);

   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity;
   glTranslatef(0,0,-Trunc(XRange*(0.9)));
   glRotatef(30,1,0,0);
   glRotatef(Rotangle,0,1,0);
   Inc(Rotangle,1);
   If RotAngle > 360
      Then Rotangle:= 0;

   glBindTexture(GL_TEXTURE_2D, Texture);
   glColor3f(175/255,120/255,64/255);
   RenderHeightMap;

   SwapBuffers(ADeviceContext);

   ErrorString:= gluErrorString(glGetError);
End;
//********************************************************************
End.
